@@ -1,28 +1,41 @@
+"""
+BitBox02
+********
+"""
+
 from typing import (
     cast,
     Any,
     Callable,
     Dict,
     Optional,
+    Mapping,
     Union,
     Tuple,
     List,
     Sequence,
     TypeVar,
 )
-from binascii import unhexlify
-import struct
 import builtins
 import sys
 from functools import wraps
 
-from ..hwwclient import HardwareWalletClient, Descriptor
-from ..serializations import (
-    PSBT,
-    CTxOut,
+import base58
+
+from ..descriptor import MultisigDescriptor
+from ..hwwclient import HardwareWalletClient
+from ..key import ExtendedKey
+from .._script import (
     is_p2pkh,
     is_p2wpkh,
     is_p2wsh,
+    parse_multisig,
+)
+from ..psbt import PSBT
+from ..tx import (
+    CTxOut,
+)
+from .._serialize import (
     ser_uint256,
     ser_sig_der,
 )
@@ -36,8 +49,16 @@ from ..errors import (
     handle_errors,
     common_err_msgs,
 )
+from ..key import (
+    KeyOriginInfo,
+    parse_path,
+)
+from ..common import (
+    AddressType,
+    Chain,
+)
 
-import hid  # type: ignore
+import hid
 
 from bitbox02 import util
 from bitbox02 import bitbox02
@@ -70,6 +91,8 @@ class BitBox02Error(UnavailableActionError):
         msg += "m/49'/0'/<account'> for p2wpkh-p2sh; "
         msg += "m/84'/0'/<account'> for p2wpkh; "
         msg += "m/48'/0'/<account'>/2' for p2wsh multisig; "
+        msg += "m/48'/0'/<account'>/1' for p2wsh-p2sh multisig; "
+        msg += "m/48'/0'/<account'>' for any supported multisig; "
         msg += "account can be between 0' and 99'; "
         msg += "For address keypaths, append /0/<address index> for a receive and /1/<change index> for a change address."
         super().__init__(msg)
@@ -133,38 +156,20 @@ class CLINoiseConfig(util.BitBoxAppNoiseConfig):
             )
 
 
-def _parse_path(nstr: str) -> Sequence[int]:
+def _keypath_hardened_prefix(keypath: Sequence[int]) -> Sequence[int]:
+    for i, e in builtins.enumerate(keypath):
+        if e & HARDENED == 0:
+            return keypath[:i]
+    return keypath
+
+
+def _xpubs_equal_ignoring_version(xpub1: bytes, xpub2: bytes) -> bool:
     """
-    Adapted from trezorlib.tools.parse_path.
-    Convert BIP32 path string to list of uint32 integers with hardened flags.
-    Several conventions are supported to set the hardened flag: -1, 1', 1h
-
-    e.g.: "0/1h/1" -> [0, 0x80000001, 1]
-
-    :param nstr: path string
-    :return: list of integers
+    Xpubs: 78 bytes. Returns true if the xpubs are equal, ignoring the 4 byte version.
+    The version is not important and allows compatibility with Electrum, which exports PSBTs with
+    xpubs using Electrum-style xpub versions.
     """
-    if not nstr:
-        return []
-
-    n = nstr.split("/")
-
-    # m/a/b/c => a/b/c
-    if n[0] == "m":
-        n = n[1:]
-
-    def str_to_harden(x: str) -> int:
-        if x.startswith("-"):
-            return abs(int(x)) + HARDENED
-        elif x.endswith(("h", "'")):
-            return int(x[:-1]) + HARDENED
-        else:
-            return int(x)
-
-    try:
-        return [str_to_harden(x) for x in n]
-    except Exception:
-        raise ValueError("Invalid BIP32 path", nstr)
+    return xpub1[4:] == xpub2[4:]
 
 
 def enumerate(password: str = "") -> List[Dict[str, object]]:
@@ -217,7 +222,7 @@ def enumerate(password: str = "") -> List[Dict[str, object]]:
                         if _using_external_gui
                         else "Please use any subcommand to unlock"
                     )
-                d_data["fingerprint"] = client.get_master_fingerprint_hex()
+                d_data["fingerprint"] = client.get_master_fingerprint().hex()
 
         result.append(d_data)
 
@@ -313,117 +318,301 @@ class Bitbox02Client(HardwareWalletClient):
     def close(self) -> None:
         self.transport.close()
 
-    def get_master_fingerprint_hex(self) -> str:
+    def get_master_fingerprint(self) -> bytes:
         """
         HWI by default retrieves the fingerprint at m/ by getting the xpub at m/0', which contains the parent fingerprint.
         The BitBox02 does not support querying arbitrary keypaths, but has an api call return the fingerprint at m/.
         """
         bb02 = self.init()
-        return bb02.root_fingerprint().hex()
+        return bb02.root_fingerprint()
 
-    def prompt_pin(self) -> Dict[str, Union[bool, str, int]]:
+    def prompt_pin(self) -> bool:
         raise UnavailableActionError(
             "The BitBox02 does not need a PIN sent from the host"
         )
 
-    def send_pin(self, pin: str) -> Dict[str, Union[bool, str, int]]:
+    def send_pin(self, pin: str) -> bool:
         raise UnavailableActionError(
             "The BitBox02 does not need a PIN sent from the host"
         )
 
     def _get_coin(self) -> bitbox02.btc.BTCCoin:
-        if self.is_testnet:
+        if self.chain != Chain.MAIN:
             return bitbox02.btc.TBTC
         return bitbox02.btc.BTC
 
     def _get_xpub(self, keypath: Sequence[int]) -> str:
         xpub_type = (
             bitbox02.btc.BTCPubRequest.TPUB
-            if self.is_testnet
+            if self.chain != Chain.MAIN
             else bitbox02.btc.BTCPubRequest.XPUB
         )
         return self.init().btc_xpub(
             keypath, coin=self._get_coin(), xpub_type=xpub_type, display=False
         )
 
-    def get_pubkey_at_path(self, bip32_path: str) -> Dict[str, str]:
-        path_uint32s = _parse_path(bip32_path)
+    def get_pubkey_at_path(self, bip32_path: str) -> ExtendedKey:
+        """
+        Fetch the public key at the derivation path.
+
+        The BitBox02 has strict keypath validation.
+
+        The only accepted keypaths for xpubs are (as of firmware v9.4.0):
+
+        - `m/49'/0'/<account'>` for `p2wpkh-p2sh` (segwit wrapped in P2SH)
+        - `m/84'/0'/<account'>` for `p2wpkh` (native segwit v0)
+        - `m/48'/0'/<account'>/2'` for p2wsh multisig (native segwit v0 multisig).
+        - `m/48'/0'/<account'>/1'` for p2wsh-p2sh multisig (p2sh-wrapped segwit v0 multisig).
+        - `m/48'/0'/<account'>` for p2wsh and p2wsh-p2sh multisig.
+
+        `account'` can be between `0'` and `99'`.
+
+        For address keypaths, append `/0/<address index>` for a receive and `/1/<change index>` for a change
+        address. Up to `10000` addresses are supported.
+
+        In testnet mode, the second element must be `1'` (e.g. `m/49'/1'/...`).
+
+        Public keys for the Legacy address type (i.e. P2WPKH and P2SH multisig) derivation path is unsupported.
+        """
+        path_uint32s = parse_path(bip32_path)
         try:
-            xpub = self._get_xpub(path_uint32s)
+            xpub_str = self._get_xpub(path_uint32s)
         except Bitbox02Exception as exc:
             raise BitBox02Error(str(exc))
-        return {"xpub": xpub}
+        xpub = ExtendedKey.deserialize(xpub_str)
+        return xpub
+
+    def _maybe_register_script_config(
+        self, script_config: bitbox02.btc.BTCScriptConfig, keypath: Sequence[int]
+    ) -> None:
+        bb02 = self.init()
+        is_registered = bb02.btc_is_script_config_registered(
+            self._get_coin(), script_config, keypath
+        )
+        if not is_registered:
+            bb02.btc_register_script_config(
+                coin=self._get_coin(),
+                script_config=script_config,
+                keypath=keypath,
+                name="",  # enter name on the device
+                xpub_type=bitbox02.btc.BTCRegisterScriptConfigRequest.AUTO_XPUB_TPUB,
+            )
+
+    def _multisig_scriptconfig(
+        self,
+        threshold: int,
+        origin_infos: Mapping[bytes, KeyOriginInfo],
+        script_type: bitbox02.btc.BTCScriptConfig.Multisig.ScriptType,
+    ) -> Tuple[bytes, bitbox02.btc.BTCScriptConfigWithKeypath]:
+        """
+        From a threshold, {xpub: KeyOriginInfo} mapping and multisig script type,
+        return our xpub and the BitBox02 multisig script config.
+        """
+        # Figure out which of the cosigners is us.
+        device_fingerprint = self.get_master_fingerprint()
+        our_xpub_index = None
+        our_account_keypath = None
+
+        xpubs: List[bytes] = []
+        for i, (xpub, keyinfo) in builtins.enumerate(origin_infos.items()):
+            xpubs.append(xpub)
+            if device_fingerprint == keyinfo.fingerprint and keyinfo.path:
+                if _xpubs_equal_ignoring_version(
+                    base58.b58decode_check(self._get_xpub(keyinfo.path)), xpub
+                ):
+                    our_xpub_index = i
+                    our_account_keypath = keyinfo.path
+
+        if our_xpub_index is None:
+            raise BadArgumentError("This BitBox02 is not one of the cosigners")
+        assert our_account_keypath
+
+        if len(xpubs) != len(set(xpubs)):
+            raise BadArgumentError("Duplicate xpubs not supported")
+
+        return (
+            xpubs[our_xpub_index],
+            bitbox02.btc.BTCScriptConfigWithKeypath(
+                script_config=bitbox02.btc.BTCScriptConfig(
+                    multisig=bitbox02.btc.BTCScriptConfig.Multisig(
+                        threshold=threshold,
+                        xpubs=[util.parse_xpub(base58.b58encode_check(xpub).decode()) for xpub in xpubs],
+                        our_xpub_index=our_xpub_index,
+                        script_type=script_type,
+                    )
+                ),
+                keypath=our_account_keypath,
+            ),
+        )
 
     @bitbox02_exception
-    def display_address(
+    def display_singlesig_address(
         self,
         bip32_path: str,
-        p2sh_p2wpkh: bool,
-        bech32: bool,
-        redeem_script: Optional[str] = None,
-        descriptor: Optional[Descriptor] = None,
-    ) -> Dict[str, str]:
-        if redeem_script:
-            raise NotImplementedError("BitBox02 multisig not integrated into HWI yet")
-
-        if p2sh_p2wpkh:
+        addr_type: AddressType,
+    ) -> str:
+        if addr_type == AddressType.SH_WIT:
             script_config = bitbox02.btc.BTCScriptConfig(
                 simple_type=bitbox02.btc.BTCScriptConfig.P2WPKH_P2SH
             )
-        elif bech32:
+        elif addr_type == AddressType.WIT:
             script_config = bitbox02.btc.BTCScriptConfig(
                 simple_type=bitbox02.btc.BTCScriptConfig.P2WPKH
             )
-        else:
+        elif addr_type == AddressType.LEGACY:
             raise UnavailableActionError(
                 "The BitBox02 does not support legacy p2pkh addresses"
             )
+        else:
+            raise BadArgumentError("Unknown address type")
         address = self.init().btc_address(
-            _parse_path(bip32_path),
+            parse_path(bip32_path),
             coin=self._get_coin(),
             script_config=script_config,
             display=True,
         )
-        return {"address": address}
+        return address
 
     @bitbox02_exception
-    def sign_tx(self, psbt: PSBT) -> Dict[str, str]:
+    def display_multisig_address(
+        self,
+        addr_type: AddressType,
+        multisig: MultisigDescriptor,
+    ) -> str:
+        if not multisig.is_sorted:
+            raise BadArgumentError("BitBox02 only supports sortedmulti descriptors")
+
+        path_suffixes = set(p.deriv_path for p in multisig.pubkeys)
+        if len(path_suffixes) != 1:
+            # Path suffix refers to the path after the account-level xpub, usually /<change>/<address>.
+            # The BitBox02 currently enforces that all of them are the same.
+            raise BadArgumentError("All multisig path suffixes must be the same")
+
+        # Figure out which of the cosigners is us.
+        key_origin_infos = {}
+        keypaths = {}
+        for pk in multisig.pubkeys:
+            assert pk.extkey and pk.origin
+            key_origin_infos[pk.extkey.serialize()] = pk.origin
+            keypaths[pk.extkey.serialize()] = pk.get_full_derivation_path(0)
+
+        if addr_type == AddressType.SH_WIT:
+            script_type = bitbox02.btc.BTCScriptConfig.Multisig.P2WSH_P2SH
+        elif addr_type == AddressType.WIT:
+            script_type = bitbox02.btc.BTCScriptConfig.Multisig.P2WSH
+        else:
+            raise BadArgumentError(
+                "BitBox02 currently only supports the following multisig script types: P2WSH, P2WSH_P2SH"
+            )
+        our_xpub, script_config_with_keypath = self._multisig_scriptconfig(
+            multisig.thresh, key_origin_infos, script_type
+        )
+        script_config = script_config_with_keypath.script_config
+        account_keypath: Sequence[int] = script_config_with_keypath.keypath
+        self._maybe_register_script_config(script_config, account_keypath)
+        keypath = parse_path(keypaths[our_xpub])
+
+        bb02 = self.init()
+        address = bb02.btc_address(
+            keypath, coin=self._get_coin(), script_config=script_config, display=True
+        )
+        return address
+
+    @bitbox02_exception
+    def sign_tx(self, psbt: PSBT) -> PSBT:
+        """
+        Sign a transaction with the BitBox02.
+
+        he BitBox02 allows mixing inputs of different script types (e.g. and `p2wpkh-p2sh` `p2wpkh`), as
+        long as the keypaths use the appropriate bip44 purpose field per input (e.g. `49'` and `84'`) and
+        all account indexes are the same.
+
+        Transactions with legacy inputs are not supported.
+        """
         def find_our_key(
-            keypaths: Dict[bytes, Sequence[int]]
+            keypaths: Dict[bytes, KeyOriginInfo]
         ) -> Tuple[Optional[bytes], Optional[Sequence[int]]]:
             """
             Keypaths is a map of pubkey to hd keypath, where the first element in the keypath is the master fingerprint. We attempt to find the key which belongs to the BitBox02 by matching the fingerprint, and then matching the pubkey.
             Returns the pubkey and the keypath, without the fingerprint.
             """
-            for pubkey, keypath_with_fingerprint in keypaths.items():
-                fp, keypath = keypath_with_fingerprint[0], keypath_with_fingerprint[1:]
+            for pubkey, origin in keypaths.items():
                 # Cheap check if the key is ours.
-                if fp != master_fp:
+                if origin.fingerprint != master_fp:
                     continue
 
                 # Expensive check if the key is ours.
                 # TODO: check for fingerprint collision
                 # keypath_account = keypath[:-2]
 
-                return pubkey, keypath
+                return pubkey, origin.path
             return None, None
 
-        def get_simple_type(
-            output: CTxOut, redeem_script: bytes
-        ) -> bitbox02.btc.BTCScriptConfig.SimpleType:
+        script_configs: List[bitbox02.btc.BTCScriptConfigWithKeypath] = []
+
+        def add_script_config(
+            script_config: bitbox02.btc.BTCScriptConfigWithKeypath
+        ) -> int:
+            # Find index of script config if already added.
+            script_config_index = next(
+                (
+                    i
+                    for i, e in builtins.enumerate(script_configs)
+                    if e.SerializeToString() == script_config.SerializeToString()
+                ),
+                None,
+            )
+            if script_config_index is not None:
+                return script_config_index
+            script_configs.append(script_config)
+            return len(script_configs) - 1
+
+        def script_config_from_utxo(
+            output: CTxOut,
+            keypath: Sequence[int],
+            redeem_script: bytes,
+            witness_script: bytes,
+        ) -> bitbox02.btc.BTCScriptConfigWithKeypath:
             if is_p2pkh(output.scriptPubKey):
                 raise BadArgumentError(
                     "The BitBox02 does not support legacy p2pkh scripts"
                 )
             if is_p2wpkh(output.scriptPubKey):
-                return bitbox02.btc.BTCScriptConfig.P2WPKH
+                return bitbox02.btc.BTCScriptConfigWithKeypath(
+                    script_config=bitbox02.btc.BTCScriptConfig(
+                        simple_type=bitbox02.btc.BTCScriptConfig.P2WPKH
+                    ),
+                    keypath=_keypath_hardened_prefix(keypath),
+                )
             if output.is_p2sh() and is_p2wpkh(redeem_script):
-                return bitbox02.btc.BTCScriptConfig.P2WPKH_P2SH
-            raise BadArgumentError(
-                "Input script type not recognized of input {}.".format(input_index)
-            )
+                return bitbox02.btc.BTCScriptConfigWithKeypath(
+                    script_config=bitbox02.btc.BTCScriptConfig(
+                        simple_type=bitbox02.btc.BTCScriptConfig.P2WPKH_P2SH
+                    ),
+                    keypath=_keypath_hardened_prefix(keypath),
+                )
+            # Check for segwit multisig (p2wsh or p2wsh-p2sh).
+            is_p2wsh_p2sh = output.is_p2sh() and is_p2wsh(redeem_script)
+            if output.is_p2wsh() or is_p2wsh_p2sh:
+                multisig = parse_multisig(witness_script)
+                if multisig:
+                    threshold, _ = multisig
+                    # We assume that all xpubs in the PSBT are part of the multisig. This is okay
+                    # since the BitBox02 enforces the same script type for all inputs and
+                    # changes. If that should change, we need to find and use the subset of xpubs
+                    # corresponding to the public keys in the current multisig script.
+                    _, script_config = self._multisig_scriptconfig(
+                        threshold,
+                        psbt.xpub,
+                        bitbox02.btc.BTCScriptConfig.Multisig.P2WSH
+                        if output.is_p2wsh()
+                        else bitbox02.btc.BTCScriptConfig.Multisig.P2WSH_P2SH,
+                    )
+                    return script_config
 
-        master_fp = struct.unpack("<I", unhexlify(self.get_master_fingerprint_hex()))[0]
+            raise BadArgumentError("Input or change script type not recognized.")
+
+        master_fp = self.get_master_fingerprint()
 
         inputs: List[bitbox02.BTCInputType] = []
 
@@ -479,8 +668,6 @@ class Bitbox02Client(HardwareWalletClient):
             assert keypath is not None
             found_pubkeys.append(found_pubkey)
 
-            # TOOD: validate keypath
-
             if bip44_account is None:
                 bip44_account = keypath[2]
             elif bip44_account != keypath[2]:
@@ -488,13 +675,11 @@ class Bitbox02Client(HardwareWalletClient):
                     "The bip44 account index must be the same for all inputs and changes"
                 )
 
-            simple_type = get_simple_type(utxo, psbt_in.redeem_script)
-
-            script_config_index_map = {
-                bitbox02.btc.BTCScriptConfig.P2WPKH: 0,
-                bitbox02.btc.BTCScriptConfig.P2WPKH_P2SH: 1,
-            }
-
+            script_config_index = add_script_config(
+                script_config_from_utxo(
+                    utxo, keypath, psbt_in.redeem_script, psbt_in.witness_script
+                )
+            )
             inputs.append(
                 {
                     "prev_out_hash": ser_uint256(tx_in.prevout.hash),
@@ -502,7 +687,7 @@ class Bitbox02Client(HardwareWalletClient):
                     "prev_out_value": utxo.nValue,
                     "sequence": tx_in.nSequence,
                     "keypath": keypath,
-                    "script_config_index": script_config_index_map[simple_type],
+                    "script_config_index": script_config_index,
                     "prev_tx": {
                         "version": prevtx.nVersion,
                         "locktime": prevtx.nLockTime,
@@ -534,12 +719,16 @@ class Bitbox02Client(HardwareWalletClient):
             is_change = keypath and keypath[-2] == 1
             if is_change:
                 assert keypath is not None
-                simple_type = get_simple_type(tx_out, psbt_out.redeem_script)
+                script_config_index = add_script_config(
+                    script_config_from_utxo(
+                        tx_out, keypath, psbt_out.redeem_script, psbt_out.witness_script
+                    )
+                )
                 outputs.append(
                     bitbox02.BTCOutputInternal(
                         keypath=keypath,
                         value=tx_out.nValue,
-                        script_config_index=script_config_index_map[simple_type],
+                        script_config_index=script_config_index,
                     )
                 )
             else:
@@ -569,24 +758,17 @@ class Bitbox02Client(HardwareWalletClient):
                 )
 
         assert bip44_account is not None
+        if (
+            len(script_configs) == 1
+            and script_configs[0].script_config.WhichOneof("config") == "multisig"
+        ):
+            self._maybe_register_script_config(
+                script_configs[0].script_config, script_configs[0].keypath
+            )
 
-        bip44_network = 1 + HARDENED if self.is_testnet else 0 + HARDENED
         sigs = self.init().btc_sign(
-            bitbox02.btc.TBTC if self.is_testnet else bitbox02.btc.BTC,
-            [
-                bitbox02.btc.BTCScriptConfigWithKeypath(
-                    script_config=bitbox02.btc.BTCScriptConfig(
-                        simple_type=bitbox02.btc.BTCScriptConfig.P2WPKH
-                    ),
-                    keypath=[84 + HARDENED, bip44_network, bip44_account],
-                ),
-                bitbox02.btc.BTCScriptConfigWithKeypath(
-                    script_config=bitbox02.btc.BTCScriptConfig(
-                        simple_type=bitbox02.btc.BTCScriptConfig.P2WPKH_P2SH
-                    ),
-                    keypath=[49 + HARDENED, bip44_network, bip44_account],
-                ),
-            ],
+            self._get_coin(),
+            script_configs,
             inputs=inputs,
             outputs=outputs,
             locktime=psbt.tx.nLockTime,
@@ -598,27 +780,27 @@ class Bitbox02Client(HardwareWalletClient):
             # ser_sig_der() adds SIGHASH_ALL
             psbt_in.partial_sigs[pubkey] = ser_sig_der(r, s)
 
-        return {"psbt": psbt.serialize()}
+        return psbt
 
     def sign_message(
         self, message: Union[str, bytes], bip32_path: str
-    ) -> Dict[str, str]:
+    ) -> str:
         raise UnavailableActionError("The BitBox02 does not support 'signmessage'")
 
     @bitbox02_exception
-    def toggle_passphrase(self) -> Dict[str, Union[bool, str, int]]:
+    def toggle_passphrase(self) -> bool:
         bb02 = self.init()
         info = bb02.device_info()
         if info["mnemonic_passphrase_enabled"]:
             bb02.disable_mnemonic_passphrase()
         else:
             bb02.enable_mnemonic_passphrase()
-        return {"success": True}
+        return True
 
     @bitbox02_exception
     def setup_device(
         self, label: str = "", passphrase: str = ""
-    ) -> Dict[str, Union[bool, str, int]]:
+    ) -> bool:
         if passphrase:
             raise UnavailableActionError(
                 "Passphrase not needed when setting up a BitBox02."
@@ -629,31 +811,33 @@ class Bitbox02Client(HardwareWalletClient):
         if label:
             bb02.set_device_name(label)
         if not bb02.set_password():
-            return {"success": False}
-        return {"success": bb02.create_backup()}
+            return False
+        return bb02.create_backup()
 
     @bitbox02_exception
-    def wipe_device(self) -> Dict[str, Union[bool, str, int]]:
-        return {"success": self.init().reset()}
+    def wipe_device(self) -> bool:
+        return self.init().reset()
 
     @bitbox02_exception
     def backup_device(
         self, label: str = "", passphrase: str = ""
-    ) -> Dict[str, Union[bool, str, int]]:
+    ) -> bool:
         if label or passphrase:
             raise UnavailableActionError(
                 "Label/passphrase not needed when exporting mnemonic from the BitBox02."
             )
 
-        return {"success": self.init().show_mnemonic()}
+        self.init().show_mnemonic()
+        return True
 
     @bitbox02_exception
     def restore_device(
         self, label: str = "", word_count: int = 24
-    ) -> Dict[str, Union[bool, str, int]]:
+    ) -> bool:
         bb02 = self.init(expect_initialized=False)
 
         if label:
             bb02.set_device_name(label)
 
-        return {"success": bb02.restore_from_mnemonic()}
+        bb02.restore_from_mnemonic()
+        return True
