@@ -1,7 +1,21 @@
-# Trezor interaction script
-
-from typing import Dict, Union
+"""
+Trezor Devices
+**************
+"""
 
+from functools import wraps
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    List,
+    NoReturn,
+    Optional,
+    Sequence,
+    Tuple,
+    Union,
+)
+from ..descriptor import MultisigDescriptor
 from ..hwwclient import HardwareWalletClient
 from ..errors import (
     ActionCanceledError,
@@ -17,54 +31,69 @@ from ..errors import (
 )
 from .trezorlib.client import TrezorClient as Trezor
 from .trezorlib.debuglink import TrezorClientDebugLink
-from .trezorlib.exceptions import Cancelled
+from .trezorlib.exceptions import Cancelled, TrezorFailure
 from .trezorlib.transport import (
-    enumerate_devices,
-    get_transport,
-    TREZOR_VENDOR_IDS,
-)
-from .trezorlib.ui import (
-    echo,
-    PassphraseUI,
-    mnemonic_words,
-    PIN_CURRENT,
-    PIN_NEW,
-    PIN_CONFIRM,
-    PIN_MATRIX_DESCRIPTION,
-    prompt,
+    DEV_TREZOR1,
+    TREZORS,
+    hid,
+    udp,
+    webusb,
 )
 from .trezorlib import (
-    tools,
     btc,
     device,
 )
-from .trezorlib import messages as proto
-from ..base58 import (
+from .trezorlib import messages
+from .._base58 import (
     get_xpub_fingerprint,
     to_address,
-    xpub_main_2_test,
 )
-from ..serializations import (
-    CTxOut,
+
+from ..key import (
     ExtendedKey,
+    parse_path,
+)
+from .._script import (
     is_p2pkh,
     is_p2sh,
     is_p2wsh,
     is_witness,
+)
+from ..psbt import PSBT
+from ..tx import (
+    CTxOut,
+)
+from .._serialize import (
     ser_uint256,
 )
-from .. import bech32
+from ..common import (
+    AddressType,
+    Chain,
+)
+from .. import _bech32 as bech32
+from mnemonic import Mnemonic
 from usb1 import USBErrorNoDevice
 from types import MethodType
 
 import base64
+import getpass
 import logging
 import sys
 
 py_enumerate = enumerate # Need to use the enumerate built-in but there's another function already named that
 
+PIN_MATRIX_DESCRIPTION = """
+Use the numeric keypad to describe number positions. The layout is:
+    7 8 9
+    4 5 6
+    1 2 3
+""".strip()
+
+Device = Union[hid.HidTransport, webusb.WebUsbTransport, udp.UdpTransport]
+
+
 # Only handles up to 15 of 15
-def parse_multisig(script):
+def parse_multisig(script: bytes) -> Tuple[bool, Optional[messages.MultisigRedeemScriptType]]:
     # Get m
     m = script[0] - 80
     if m < 1 or m > 15:
@@ -81,8 +110,8 @@ def parse_multisig(script):
         key = script[offset:offset + 33]
         offset += 33
 
-        hd_node = proto.HDNodeType(depth=0, fingerprint=0, child_num=0, chain_code=b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00', public_key=key)
-        pubkeys.append(proto.HDNodePathType(node=hd_node, address_n=[]))
+        hd_node = messages.HDNodeType(depth=0, fingerprint=0, child_num=0, chain_code=b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00', public_key=key)
+        pubkeys.append(messages.HDNodePathType(node=hd_node, address_n=[]))
 
     # Check things at the end
     n = script[offset] - 80
@@ -94,11 +123,13 @@ def parse_multisig(script):
         return (False, None)
 
     # Build MultisigRedeemScriptType and return it
-    multisig = proto.MultisigRedeemScriptType(m=m, signatures=[b''] * n, pubkeys=pubkeys)
+    multisig = messages.MultisigRedeemScriptType(m=m, signatures=[b''] * n, pubkeys=pubkeys)
     return (True, multisig)
 
-def trezor_exception(f):
-    def func(*args, **kwargs):
+
+def trezor_exception(f: Callable[..., Any]) -> Any:
+    @wraps(f)
+    def func(*args: Any, **kwargs: Any) -> Any:
         try:
             return f(*args, **kwargs)
         except ValueError as e:
@@ -109,39 +140,111 @@ def trezor_exception(f):
             raise DeviceConnectionError('Device disconnected')
     return func
 
-def interactive_get_pin(self, code=None):
-    if code == PIN_CURRENT:
+
+def interactive_get_pin(self: object, code: Optional[int] = None) -> str:
+    if code == messages.PinMatrixRequestType.Currrent:
         desc = "current PIN"
-    elif code == PIN_NEW:
+    elif code == messages.PinMatrixRequestType.NewFirst:
         desc = "new PIN"
-    elif code == PIN_CONFIRM:
+    elif code == messages.PinMatrixRequestType.NewSecond:
         desc = "new PIN again"
     else:
         desc = "PIN"
 
-    echo(PIN_MATRIX_DESCRIPTION)
+    print(PIN_MATRIX_DESCRIPTION, file=sys.stderr)
 
     while True:
-        pin = prompt("Please enter {}".format(desc), hide_input=True)
+        pin = getpass.getpass(f"Please entire {desc}:\n")
         if not pin.isdigit():
-            echo("Non-numerical PIN provided, please try again")
+            print("Non-numerical PIN provided, please try again", file=sys.stderr)
         else:
             return pin
 
+
+def mnemonic_words(expand: bool = False, language: str = "english") -> Callable[[Any], str]:
+    wordlist: Sequence[str] = []
+    if expand:
+        wordlist = Mnemonic(language).wordlist
+
+    def expand_word(word: str) -> str:
+        if not expand:
+            return word
+        if word in wordlist:
+            return word
+        matches = [w for w in wordlist if w.startswith(word)]
+        if len(matches) == 1:
+            return matches[0]
+        print("Choose one of: " + ", ".join(matches), file=sys.stderr)
+        raise KeyError(word)
+
+    def get_word(type: messages.WordRequestType) -> str:
+        assert type == messages.WordRequestType.Plain
+        while True:
+            try:
+                word = input("Enter one word of mnemonic:\n")
+                return expand_word(word)
+            except KeyError:
+                pass
+            except Exception:
+                raise Cancelled from None
+
+    return get_word
+
+
+class PassphraseUI:
+    def __init__(self, passphrase: str) -> None:
+        self.passphrase = passphrase
+        self.pinmatrix_shown = False
+        self.prompt_shown = False
+        self.always_prompt = False
+        self.return_passphrase = True
+
+    def button_request(self, code: Optional[int]) -> None:
+        if not self.prompt_shown:
+            print("Please confirm action on your Trezor device", file=sys.stderr)
+        if not self.always_prompt:
+            self.prompt_shown = True
+
+    def get_pin(self, code: Optional[int] = None) -> NoReturn:
+        raise NotImplementedError('get_pin is not needed')
+
+    def disallow_passphrase(self) -> None:
+        self.return_passphrase = False
+
+    def get_passphrase(self, available_on_device: bool) -> str:
+        if self.return_passphrase:
+            return self.passphrase
+        raise ValueError('Passphrase from Host is not allowed for Trezor T')
+
+
+HID_IDS = {DEV_TREZOR1}
+WEBUSB_IDS = TREZORS.copy()
+
+
+def get_path_transport(path: str) -> Device:
+    devs = hid.HidTransport.enumerate(usb_ids=HID_IDS)
+    devs.extend(webusb.WebUsbTransport.enumerate(usb_ids=WEBUSB_IDS))
+    devs.extend(udp.UdpTransport.enumerate())
+    for dev in devs:
+        if path == dev.get_path():
+            return dev
+    raise BadArgumentError(f"Could not find device by path: {path}")
+
+
 # This class extends the HardwareWalletClient for Trezor specific things
 class TrezorClient(HardwareWalletClient):
 
-    def __init__(self, path, password='', expert=False):
+    def __init__(self, path: str, password: str = "", expert: bool = False) -> None:
         super(TrezorClient, self).__init__(path, password, expert)
         self.simulator = False
+        transport = get_path_transport(path)
         if path.startswith('udp'):
             logging.debug('Simulator found, using DebugLink')
-            transport = get_transport(path)
             self.client = TrezorClientDebugLink(transport=transport)
             self.simulator = True
-            self.client.set_passphrase(password)
+            self.client.use_passphrase(password)
         else:
-            self.client = Trezor(transport=get_transport(path), ui=PassphraseUI(password))
+            self.client = Trezor(transport=transport, ui=PassphraseUI(password))
 
         # if it wasn't able to find a client, throw an error
         if not self.client:
@@ -150,38 +253,48 @@ class TrezorClient(HardwareWalletClient):
         self.password = password
         self.type = 'Trezor'
 
-    def _check_unlocked(self):
-        self.coin_name = 'Testnet' if self.is_testnet else 'Bitcoin'
-        self.client.init_device()
-        if self.client.features.model == 'T':
+    def _prepare_device(self) -> None:
+        self.coin_name = 'Bitcoin' if self.chain == Chain.MAIN else 'Testnet'
+        resp = self.client.refresh_features()
+        # If this is a Trezor One or Keepkey, do Initialize
+        if resp.model == '1' or resp.model == 'K1-14AM':
+            self.client.init_device()
+        # For the T, we need to check if a passphrase needs to be entered
+        elif resp.model == 'T':
+            try:
+                self.client.ensure_unlocked()
+            except TrezorFailure:
+                self.client.init_device()
+
+    def _check_unlocked(self) -> None:
+        self._prepare_device()
+        if self.client.features.model == 'T' and isinstance(self.client.ui, PassphraseUI):
             self.client.ui.disallow_passphrase()
-        if self.client.features.pin_protection and not self.client.features.pin_cached:
+        if self.client.features.pin_protection and not self.client.features.unlocked:
             raise DeviceNotReadyError('{} is locked. Unlock by using \'promptpin\' and then \'sendpin\'.'.format(self.type))
 
-    # Must return a dict with the xpub
-    # Retrieves the public key at the specified BIP 32 derivation path
     @trezor_exception
-    def get_pubkey_at_path(self, path):
+    def get_pubkey_at_path(self, path: str) -> ExtendedKey:
         self._check_unlocked()
         try:
-            expanded_path = tools.parse_path(path)
+            expanded_path = parse_path(path)
         except ValueError as e:
             raise BadArgumentError(str(e))
         output = btc.get_public_node(self.client, expanded_path, coin_name=self.coin_name)
-        if self.is_testnet:
-            result = {'xpub': xpub_main_2_test(output.xpub)}
-        else:
-            result = {'xpub': output.xpub}
-        if self.expert:
-            xpub_obj = ExtendedKey()
-            xpub_obj.deserialize(output.xpub)
-            result.update(xpub_obj.get_printable_dict())
-        return result
+        xpub = ExtendedKey.deserialize(output.xpub)
+        if self.chain != Chain.MAIN:
+            xpub.version = ExtendedKey.TESTNET_PUBLIC
+        return xpub
 
-    # Must return a hex string with the signed transaction
-    # The tx must be in the psbt format
     @trezor_exception
-    def sign_tx(self, tx):
+    def sign_tx(self, tx: PSBT) -> PSBT:
+        """
+        Sign a transaction with the Trezor. There are some limitations to what transactions can be signed.
+
+        - Multisig inputs are limited to at most n-of-15 multisigs. This is a firmware limitation.
+        - Transactions with arbitrary input scripts (scriptPubKey, redeemScript, or witnessScript) and arbitrary output scripts cannot be signed. This is a firmware limitation.
+        - Send-to-self transactions will result in no prompt for outputs as all outputs will be detected as change.
+        """
         self._check_unlocked()
 
         # Get this devices master key fingerprint
@@ -197,12 +310,11 @@ class TrezorClient(HardwareWalletClient):
             inputs = []
             to_ignore = [] # Note down which inputs whose signatures we're going to ignore
             for input_num, (psbt_in, txin) in py_enumerate(list(zip(tx.inputs, tx.tx.vin))):
-                txinputtype = proto.TxInputType()
-
-                # Set the input stuff
-                txinputtype.prev_hash = ser_uint256(txin.prevout.hash)[::-1]
-                txinputtype.prev_index = txin.prevout.n
-                txinputtype.sequence = txin.nSequence
+                txinputtype = messages.TxInputType(
+                    prev_hash=ser_uint256(txin.prevout.hash)[::-1],
+                    prev_index=txin.prevout.n,
+                    sequence=txin.nSequence,
+                )
 
                 # Detrermine spend type
                 scriptcode = b''
@@ -231,11 +343,11 @@ class TrezorClient(HardwareWalletClient):
 
                 if is_wit:
                     if p2sh:
-                        txinputtype.script_type = proto.InputScriptType.SPENDP2SHWITNESS
+                        txinputtype.script_type = messages.InputScriptType.SPENDP2SHWITNESS
                     else:
-                        txinputtype.script_type = proto.InputScriptType.SPENDWITNESS
+                        txinputtype.script_type = messages.InputScriptType.SPENDWITNESS
                 else:
-                    txinputtype.script_type = proto.InputScriptType.SPENDADDRESS
+                    txinputtype.script_type = messages.InputScriptType.SPENDADDRESS
                 txinputtype.amount = utxo.nValue
 
                 # Check if P2WSH
@@ -247,10 +359,10 @@ class TrezorClient(HardwareWalletClient):
                     scriptcode = psbt_in.witness_script
                     p2wsh = True
 
-                def ignore_input():
-                    txinputtype.address_n = [0x80000000 | 84, 0x80000000 | (1 if self.is_testnet else 0)]
+                def ignore_input() -> None:
+                    txinputtype.address_n = [0x80000000 | 84, 0x80000000 | (0 if self.chain == Chain.MAIN else 1), 0x80000000, 0, 0]
                     txinputtype.multisig = None
-                    txinputtype.script_type = proto.InputScriptType.SPENDWITNESS
+                    txinputtype.script_type = messages.InputScriptType.SPENDWITNESS
                     inputs.append(txinputtype)
                     to_ignore.append(input_num)
 
@@ -261,7 +373,7 @@ class TrezorClient(HardwareWalletClient):
                     txinputtype.multisig = multisig
                     if not is_wit:
                         if utxo.is_p2sh:
-                            txinputtype.script_type = proto.InputScriptType.SPENDMULTISIG
+                            txinputtype.script_type = messages.InputScriptType.SPENDMULTISIG
                         else:
                             # Cannot sign bare multisig, ignore it
                             ignore_input()
@@ -281,12 +393,12 @@ class TrezorClient(HardwareWalletClient):
                 our_keys = 0
                 for key in psbt_in.hd_keypaths.keys():
                     keypath = psbt_in.hd_keypaths[key]
-                    if keypath[0] == master_fp:
+                    if keypath.fingerprint == master_fp:
                         if key in psbt_in.partial_sigs: # This key already has a signature
                             found_in_sigs = True
                             continue
                         if not found: # This key does not have a signature and we don't have a key to sign with yet
-                            txinputtype.address_n = keypath[1:]
+                            txinputtype.address_n = keypath.path
                             found = True
                         our_keys += 1
 
@@ -306,7 +418,7 @@ class TrezorClient(HardwareWalletClient):
                 inputs.append(txinputtype)
 
             # address version byte
-            if self.is_testnet:
+            if self.chain != Chain.MAIN:
                 p2pkh_version = b'\x6f'
                 p2sh_version = b'\xc4'
                 bech32_hrp = 'tb'
@@ -318,13 +430,15 @@ class TrezorClient(HardwareWalletClient):
             # prepare outputs
             outputs = []
             for i, out in py_enumerate(tx.tx.vout):
-                txoutput = proto.TxOutputType()
-                txoutput.amount = out.nValue
-                txoutput.script_type = proto.OutputScriptType.PAYTOADDRESS
+                txoutput = messages.TxOutputType(amount=out.nValue)
+                txoutput.script_type = messages.OutputScriptType.PAYTOADDRESS
                 if out.is_p2pkh():
                     txoutput.address = to_address(out.scriptPubKey[3:23], p2pkh_version)
                 elif out.is_p2sh():
                     txoutput.address = to_address(out.scriptPubKey[2:22], p2sh_version)
+                elif out.is_opreturn():
+                    txoutput.script_type = messages.OutputScriptType.PAYTOOPRETURN
+                    txoutput.op_return_data = out.scriptPubKey[2:]
                 else:
                     wit, ver, prog = out.is_witness()
                     if wit:
@@ -336,20 +450,20 @@ class TrezorClient(HardwareWalletClient):
                 psbt_out = tx.outputs[i]
                 if len(psbt_out.hd_keypaths) == 1:
                     _, keypath = next(iter(psbt_out.hd_keypaths.items()))
-                    if keypath[0] == master_fp:
+                    if keypath.fingerprint == master_fp:
                         wit, ver, prog = out.is_witness()
                         if out.is_p2pkh():
-                            txoutput.address_n = keypath[1:]
+                            txoutput.address_n = keypath.path
                             txoutput.address = None
                         elif wit:
-                            txoutput.script_type = proto.OutputScriptType.PAYTOWITNESS
-                            txoutput.address_n = keypath[1:]
+                            txoutput.script_type = messages.OutputScriptType.PAYTOWITNESS
+                            txoutput.address_n = keypath.path
                             txoutput.address = None
                         elif out.is_p2sh() and psbt_out.redeem_script:
                             wit, ver, prog = CTxOut(0, psbt_out.redeem_script).is_witness()
                             if wit and len(prog) == 20:
-                                txoutput.script_type = proto.OutputScriptType.PAYTOP2SHWITNESS
-                                txoutput.address_n = keypath[1:]
+                                txoutput.script_type = messages.OutputScriptType.PAYTOP2SHWITNESS
+                                txoutput.address_n = keypath.path
                                 txoutput.address = None
 
                 # append to outputs
@@ -361,114 +475,154 @@ class TrezorClient(HardwareWalletClient):
                 if psbt_in.non_witness_utxo:
                     prev = psbt_in.non_witness_utxo
 
-                    t = proto.TransactionType()
+                    t = messages.TransactionType()
                     t.version = prev.nVersion
                     t.lock_time = prev.nLockTime
 
                     for vin in prev.vin:
-                        i = proto.TxInputType()
-                        i.prev_hash = ser_uint256(vin.prevout.hash)[::-1]
-                        i.prev_index = vin.prevout.n
-                        i.script_sig = vin.scriptSig
-                        i.sequence = vin.nSequence
+                        i = messages.TxInputType(
+                            prev_hash=ser_uint256(vin.prevout.hash)[::-1],
+                            prev_index=vin.prevout.n,
+                            script_sig=vin.scriptSig,
+                            sequence=vin.nSequence,
+                        )
                         t.inputs.append(i)
 
                     for vout in prev.vout:
-                        o = proto.TxOutputBinType()
-                        o.amount = vout.nValue
-                        o.script_pubkey = vout.scriptPubKey
+                        o = messages.TxOutputBinType(
+                            amount=vout.nValue,
+                            script_pubkey=vout.scriptPubKey,
+                        )
                         t.bin_outputs.append(o)
                     logging.debug(psbt_in.non_witness_utxo.hash)
+                    assert psbt_in.non_witness_utxo.sha256 is not None
                     prevtxs[ser_uint256(psbt_in.non_witness_utxo.sha256)[::-1]] = t
 
             # Sign the transaction
-            tx_details = proto.SignTx()
-            tx_details.version = tx.tx.nVersion
-            tx_details.lock_time = tx.tx.nLockTime
-            signed_tx = btc.sign_tx(self.client, self.coin_name, inputs, outputs, tx_details, prevtxs)
+            signed_tx = btc.sign_tx(
+                client=self.client,
+                coin_name=self.coin_name,
+                inputs=inputs,
+                outputs=outputs,
+                prev_txes=prevtxs,
+                version=tx.tx.nVersion,
+                lock_time=tx.tx.nLockTime,
+            )
 
             # Each input has one signature
             for input_num, (psbt_in, sig) in py_enumerate(list(zip(tx.inputs, signed_tx[0]))):
                 if input_num in to_ignore:
                     continue
                 for pubkey in psbt_in.hd_keypaths.keys():
-                    fp = psbt_in.hd_keypaths[pubkey][0]
+                    fp = psbt_in.hd_keypaths[pubkey].fingerprint
                     if fp == master_fp and pubkey not in psbt_in.partial_sigs:
                         psbt_in.partial_sigs[pubkey] = sig + b'\x01'
                         break
 
             p += 1
 
-        return {'psbt': tx.serialize()}
+        return tx
 
     @trezor_exception
-    def sign_message(self, message: Union[str, bytes], keypath: str) -> Dict[str, str]:
+    def sign_message(self, message: Union[str, bytes], keypath: str) -> str:
         self._check_unlocked()
-        path = tools.parse_path(keypath)
+        path = parse_path(keypath)
         result = btc.sign_message(self.client, self.coin_name, path, message)
-        return {'signature': base64.b64encode(result.signature).decode('utf-8')}
+        return base64.b64encode(result.signature).decode('utf-8')
 
-    # Display address of specified type on the device.
     @trezor_exception
-    def display_address(self, keypath, p2sh_p2wpkh, bech32, redeem_script=None, descriptor=None):
+    def display_singlesig_address(
+        self,
+        keypath: str,
+        addr_type: AddressType,
+    ) -> str:
         self._check_unlocked()
 
-        # descriptor means multisig with xpubs
-        if descriptor:
-            pubkeys = []
-            xpub = ExtendedKey()
-            for i in range(0, descriptor.multisig_N):
-                xpub.deserialize(descriptor.base_key[i])
-                hd_node = proto.HDNodeType(depth=xpub.depth, fingerprint=int.from_bytes(xpub.parent_fingerprint, 'big'), child_num=xpub.child_num, chain_code=xpub.chaincode, public_key=xpub.pubkey)
-                pubkeys.append(proto.HDNodePathType(node=hd_node, address_n=tools.parse_path('m' + descriptor.path_suffix[i])))
-            multisig = proto.MultisigRedeemScriptType(m=int(descriptor.multisig_M), signatures=[b''] * int(descriptor.multisig_N), pubkeys=pubkeys)
-        # redeem_script means p2sh/multisig
-        elif redeem_script:
-            # Get multisig object required by Trezor's get_address
-            multisig = parse_multisig(bytes.fromhex(redeem_script))
-            if not multisig[0]:
-                raise BadArgumentError("The redeem script provided is not a multisig. Only multisig scripts can be displayed.")
-            multisig = multisig[1]
-        else:
-            multisig = None
-
         # Script type
-        if p2sh_p2wpkh:
-            script_type = proto.InputScriptType.SPENDP2SHWITNESS
-        elif bech32:
-            script_type = proto.InputScriptType.SPENDWITNESS
-        elif redeem_script:
-            script_type = proto.InputScriptType.SPENDMULTISIG
+        if addr_type == AddressType.SH_WIT:
+            script_type = messages.InputScriptType.SPENDP2SHWITNESS
+        elif addr_type == AddressType.WIT:
+            script_type = messages.InputScriptType.SPENDWITNESS
+        elif addr_type == AddressType.LEGACY:
+            script_type = messages.InputScriptType.SPENDADDRESS
         else:
-            script_type = proto.InputScriptType.SPENDADDRESS
+            raise BadArgumentError("Unknown address type")
 
-        # convert device fingerprint to 'm' if exists in path
-        keypath = keypath.replace(self.get_master_fingerprint_hex(), 'm')
+        expanded_path = parse_path(keypath)
 
-        for path in keypath.split(','):
-            if len(path.split('/')[0]) == 8:
-                path = path.split('/', 1)[1]
-            expanded_path = tools.parse_path(path)
+        try:
+            address = btc.get_address(
+                self.client,
+                self.coin_name,
+                expanded_path,
+                show_display=True,
+                script_type=script_type,
+                multisig=None,
+            )
+            assert isinstance(address, str)
+            return address
+        except Exception:
+            pass
 
+        raise BadArgumentError("No path supplied matched device keys")
+
+    @trezor_exception
+    def display_multisig_address(
+        self,
+        addr_type: AddressType,
+        multisig: MultisigDescriptor,
+    ) -> str:
+        self._check_unlocked()
+
+        der_pks = list(zip([p.get_pubkey_bytes(0) for p in multisig.pubkeys], multisig.pubkeys))
+        if multisig.is_sorted:
+            der_pks = sorted(der_pks)
+
+        pubkey_objs = []
+        for pk, p in der_pks:
+            if p.extkey is not None:
+                xpub = p.extkey
+                hd_node = messages.HDNodeType(depth=xpub.depth, fingerprint=int.from_bytes(xpub.parent_fingerprint, 'big'), child_num=xpub.child_num, chain_code=xpub.chaincode, public_key=xpub.pubkey)
+                pubkey_objs.append(messages.HDNodePathType(node=hd_node, address_n=parse_path("m" + p.deriv_path if p.deriv_path is not None else "")))
+            else:
+                hd_node = messages.HDNodeType(depth=0, fingerprint=0, child_num=0, chain_code=b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00', public_key=pk)
+                pubkey_objs.append(messages.HDNodePathType(node=hd_node, address_n=[]))
+
+        trezor_ms = messages.MultisigRedeemScriptType(m=multisig.thresh, signatures=[b''] * len(pubkey_objs), pubkeys=pubkey_objs)
+
+        # Script type
+        if addr_type == AddressType.SH_WIT:
+            script_type = messages.InputScriptType.SPENDP2SHWITNESS
+        elif addr_type == AddressType.WIT:
+            script_type = messages.InputScriptType.SPENDWITNESS
+        elif addr_type == AddressType.LEGACY:
+            script_type = messages.InputScriptType.SPENDMULTISIG
+        else:
+            raise BadArgumentError("Unknown address type")
+
+        for p in multisig.pubkeys:
+            keypath = p.origin.get_derivation_path() if p.origin is not None else "m/"
+            keypath += p.deriv_path if p.deriv_path is not None else ""
+            path = parse_path(keypath)
             try:
                 address = btc.get_address(
                     self.client,
                     self.coin_name,
-                    expanded_path,
+                    path,
                     show_display=True,
                     script_type=script_type,
-                    multisig=multisig,
+                    multisig=trezor_ms,
                 )
-                return {'address': address}
+                assert isinstance(address, str)
+                return address
             except Exception:
                 pass
 
         raise BadArgumentError("No path supplied matched device keys")
 
-    # Setup a new device
     @trezor_exception
-    def setup_device(self, label='', passphrase=''):
-        self.client.init_device()
+    def setup_device(self, label: str = "", passphrase: str = "") -> bool:
+        self._prepare_device()
         if not self.simulator:
             # Use interactive_get_pin
             self.client.ui.get_pin = MethodType(interactive_get_pin, self.client.ui)
@@ -476,70 +630,72 @@ class TrezorClient(HardwareWalletClient):
         if self.client.features.initialized:
             raise DeviceAlreadyInitError('Device is already initialized. Use wipe first and try again')
         device.reset(self.client, passphrase_protection=bool(self.password))
-        return {'success': True}
+        return True
 
-    # Wipe this device
     @trezor_exception
-    def wipe_device(self):
+    def wipe_device(self) -> bool:
         self._check_unlocked()
         device.wipe(self.client)
-        return {'success': True}
+        return True
 
-    # Restore device from mnemonic or xprv
     @trezor_exception
-    def restore_device(self, label='', word_count=24):
-        self.client.init_device()
+    def restore_device(self, label: str = "", word_count: int = 24) -> bool:
+        self._prepare_device()
         if not self.simulator:
             # Use interactive_get_pin
             self.client.ui.get_pin = MethodType(interactive_get_pin, self.client.ui)
 
         device.recover(self.client, word_count=word_count, label=label, input_callback=mnemonic_words(), passphrase_protection=bool(self.password))
-        return {'success': True}
+        return True
 
-    # Begin backup process
-    def backup_device(self, label='', passphrase=''):
+    def backup_device(self, label: str = "", passphrase: str = "") -> bool:
+        """
+        Trezor devices do not support backing up via software.
+
+        :raises UnavailableActionError: Always, this function is unavailable
+        """
         raise UnavailableActionError('The {} does not support creating a backup via software'.format(self.type))
 
-    # Close the device
     @trezor_exception
-    def close(self):
+    def close(self) -> None:
         self.client.close()
 
-    # Prompt for a pin on device
     @trezor_exception
-    def prompt_pin(self):
-        self.coin_name = 'Testnet' if self.is_testnet else 'Bitcoin'
+    def prompt_pin(self) -> bool:
+        self.coin_name = 'Bitcoin' if self.chain == Chain.MAIN else 'Testnet'
         self.client.open()
-        self.client.init_device()
+        self._prepare_device()
         if not self.client.features.pin_protection:
             raise DeviceAlreadyUnlockedError('This device does not need a PIN')
-        if self.client.features.pin_cached:
+        if self.client.features.unlocked:
             raise DeviceAlreadyUnlockedError('The PIN has already been sent to this device')
         print('Use \'sendpin\' to provide the number positions for the PIN as displayed on your device\'s screen', file=sys.stderr)
         print(PIN_MATRIX_DESCRIPTION, file=sys.stderr)
-        self.client.call_raw(proto.GetPublicKey(address_n=[0x8000002c, 0x80000001, 0x80000000], ecdsa_curve_name=None, show_display=False, coin_name=self.coin_name, script_type=proto.InputScriptType.SPENDADDRESS))
-        return {'success': True}
+        self.client.call_raw(messages.GetPublicKey(address_n=[0x8000002c, 0x80000001, 0x80000000], ecdsa_curve_name=None, show_display=False, coin_name=self.coin_name, script_type=messages.InputScriptType.SPENDADDRESS))
+        return True
 
-    # Send the pin
     @trezor_exception
-    def send_pin(self, pin):
+    def send_pin(self, pin: str) -> bool:
         self.client.open()
         if not pin.isdigit():
             raise BadArgumentError("Non-numeric PIN provided")
-        resp = self.client.call_raw(proto.PinMatrixAck(pin=pin))
-        if isinstance(resp, proto.Failure):
-            self.client.features = self.client.call_raw(proto.GetFeatures())
-            if isinstance(self.client.features, proto.Features):
+        resp = self.client.call_raw(messages.PinMatrixAck(pin=pin))
+        if isinstance(resp, messages.Failure):
+            self.client.features = self.client.call_raw(messages.GetFeatures())
+            if isinstance(self.client.features, messages.Features):
                 if not self.client.features.pin_protection:
                     raise DeviceAlreadyUnlockedError('This device does not need a PIN')
-                if self.client.features.pin_cached:
+                if self.client.features.unlocked:
                     raise DeviceAlreadyUnlockedError('The PIN has already been sent to this device')
-            return {'success': False}
-        return {'success': True}
+            return False
+        elif isinstance(resp, messages.PassphraseRequest):
+            pass_resp = self.client.call_raw(messages.PassphraseAck(passphrase=self.client.ui.get_passphrase(available_on_device=False), on_device=False))
+            if isinstance(pass_resp, messages.Deprecated_PassphraseStateRequest):
+                self.client.call_raw(messages.Deprecated_PassphraseStateAck())
+        return True
 
-    # Toggle passphrase
     @trezor_exception
-    def toggle_passphrase(self):
+    def toggle_passphrase(self) -> bool:
         self._check_unlocked()
         try:
             device.apply_settings(self.client, use_passphrase=not self.client.features.passphrase_protection)
@@ -548,16 +704,15 @@ class TrezorClient(HardwareWalletClient):
                 print('Confirm the action by entering your PIN', file=sys.stderr)
                 print('Use \'sendpin\' to provide the number positions for the PIN as displayed on your device\'s screen', file=sys.stderr)
                 print(PIN_MATRIX_DESCRIPTION, file=sys.stderr)
-        return {'success': True}
+        return True
 
-def enumerate(password=''):
+def enumerate(password: str = "") -> List[Dict[str, Any]]:
     results = []
-    for dev in enumerate_devices():
-        # enumerate_devices filters to Trezors and Keepkeys.
-        # Only allow Trezors and unknowns. Unknown devices will reach the check for vendor later
-        if dev.get_usb_vendor_id() not in TREZOR_VENDOR_IDS | {-1}:
-            continue
-        d_data = {}
+    devs = hid.HidTransport.enumerate()
+    devs.extend(webusb.WebUsbTransport.enumerate())
+    devs.extend(udp.UdpTransport.enumerate())
+    for dev in devs:
+        d_data: Dict[str, Any] = {}
 
         d_data['type'] = 'trezor'
         d_data['path'] = dev.get_path()
@@ -565,7 +720,10 @@ def enumerate(password=''):
         client = None
         with handle_errors(common_err_msgs["enumerate"], d_data):
             client = TrezorClient(d_data['path'], password)
-            client.client.init_device()
+            try:
+                client._prepare_device()
+            except TypeError:
+                continue
             if 'trezor' not in client.client.features.vendor:
                 continue
 
@@ -573,7 +731,7 @@ def enumerate(password=''):
             if d_data['path'] == 'udp:127.0.0.1:21324':
                 d_data['model'] += '_simulator'
 
-            d_data['needs_pin_sent'] = client.client.features.pin_protection and not client.client.features.pin_cached
+            d_data['needs_pin_sent'] = client.client.features.pin_protection and not client.client.features.unlocked
             if client.client.features.model == '1':
                 d_data['needs_passphrase_sent'] = client.client.features.passphrase_protection # always need the passphrase sent for Trezor One if it has passphrase protection enabled
             else:
@@ -583,7 +741,7 @@ def enumerate(password=''):
             if d_data['needs_passphrase_sent'] and not password:
                 raise DeviceNotReadyError("Passphrase needs to be specified before the fingerprint information can be retrieved")
             if client.client.features.initialized:
-                d_data['fingerprint'] = client.get_master_fingerprint_hex()
+                d_data['fingerprint'] = client.get_master_fingerprint().hex()
                 d_data['needs_passphrase_sent'] = False # Passphrase is always needed for the above to have worked, so it's already sent
             else:
                 d_data['error'] = 'Not initialized'
